@@ -8,7 +8,7 @@
  work, as well as functions which operate on them.
 """
 from types import FunctionType as Function
-from typing import Any, Dict, Generic, TypeVar
+from typing import Any, Dict, Generic, Hashable, TypeVar
 
 DataClass = Any  # type hint for variables that should be data class instances
 T = TypeVar('T')
@@ -26,7 +26,7 @@ class Internal(Generic[T]):
 class DataClassMeta(type):
     """The metaclass for a data class."""
     DEFAULT_OPTIONS = dict(init=True, repr=True, eq=True, iter=False, frozen=False, kwargs=False, slots=False,
-                           hide_internals=True)
+                           order=False, unsafe_hash=True, hide_internals=True)
 
     def __new__(mcs, name, bases, dict_, **kwargs):
         # collect annotations, defaults, slots and options from this class' ancestors, in definition order
@@ -52,11 +52,9 @@ class DataClassMeta(type):
         # delete what will become stale references so that Python creates new ones
         del dict_['__dict__'], dict_['__weakref__']
 
-        # warn the user if they try to use __post_init__
-        if '__post_init__' in dict_:
-            raise TypeError('dataclassy does not use __post_init__. You should rename this method __init__')
-
         # create/apply generated methods and attributes
+        user_init = type(dict_.get('__init__')) is Function
+
         if options['slots']:
             # values with default values must only be present in slots, not dict, otherwise Python will interpret them
             # as read only
@@ -69,21 +67,36 @@ class DataClassMeta(type):
                 del dict_[d]
             del dict_['__slots__']
         if options['init']:
-            dict_.setdefault('__new__', _generate_new(all_annotations, all_defaults,
+            dict_.setdefault('__new__', _generate_new(all_annotations, all_defaults, user_init,
                                                       options['kwargs'], options['frozen']))
         if options['repr']:
-            dict_.setdefault('__repr__', _generated_repr)
+            dict_.setdefault('__repr__', __repr__)
         if options['eq']:
-            dict_.setdefault('__eq__', _generated_eq)
+            dict_.setdefault('__eq__', __eq__)
         if options['iter']:
-            dict_.setdefault('__iter__', _generated_iter)
+            dict_.setdefault('__iter__', __iter__)
         if options['frozen']:
-            dict_['__delattr__'] = dict_['__setattr__'] = _generated_attr
+            dict_['__delattr__'] = dict_['__setattr__'] = __setattr__
+        if options['order']:
+            dict_.setdefault('__lt__', __lt__)
+        if (options['eq'] and options['frozen']) or options['unsafe_hash']:
+            dict_.setdefault('__hash__', __hash__)
 
-        return type.__new__(mcs, name, bases, dict_)
+        return type.__new__(mcs if not user_init else DataClassInit, name, bases, dict_)
 
+    def __init__(cls, *args, **kwargs):
+        # warn the user if they try to use __post_init__
+        if hasattr(cls, '__post_init__'):
+            raise TypeError('dataclassy does not use __post_init__. You should rename this method __init__')
+
+        if cls.__dataclass__['eq'] and cls.__dataclass__['order']:
+            from functools import total_ordering
+            total_ordering(cls)
+
+
+class DataClassInit(DataClassMeta):
+    """In the case that a custom __init__ is defined, remove arguments used by __new__ before calling it."""
     def __call__(cls, *args, **kwargs):
-        """Remove arguments used by __new__ before calling __init__."""
         instance = cls.__new__(cls, *args, **kwargs)
 
         args = args[cls.__new__.__code__.co_argcount - 1:]  # -1 for 'cls'
@@ -101,13 +114,10 @@ class DataClassMeta(type):
         return inspect.Signature(parameters[1:])  # remove 'cls' to transform parameters of __new__ into those of class
 
 
-def _generate_new(annotations: Dict[str, Any], defaults: Dict[str, Any], gen_kwargs: bool, frozen: bool) -> Function:
+def _generate_new(annotations: Dict, defaults: Dict, user_init: bool, gen_kwargs: bool, frozen: bool) -> Function:
     """Generate and return a __new__ method for a data class which has as parameters all fields of the data class.
     When the data class is initialised, arguments to this function are applied to the fields of the new instance. Using
     __new__ frees up __init__, allowing it to be defined by the user to perform additional, custom initialisation."""
-    user_init = '__init__' in defaults
-
-    # determine arguments for initialiser
     arguments = [a for a in annotations if a not in defaults]
     default_arguments = [f'{a}={a}' for a in annotations if a in defaults]
     args = ['*args'] if user_init else []  # if init is defined, new's arguments must be kw-only to avoid ambiguity
@@ -133,11 +143,34 @@ def _generate_new(annotations: Dict[str, Any], defaults: Dict[str, Any], gen_kwa
     return function
 
 
-# method implementations which are constant for all data classes
-from .functions import fields, as_tuple, is_dataclass_instance
+# generic method implementations common to all data classes
+# these are currently relatively inefficient - it would be better to cache an expression for a class' tuple
+from .functions import values, as_tuple
 
-_generated_eq = lambda self, other: is_dataclass_instance(other) and as_tuple(self) == as_tuple(other)
-_generated_iter = lambda self: iter(as_tuple(self))
-_generated_repr = lambda self: f'{self.__class__.__name__}(' \
-        f'{", ".join(f"{f}={v!r}" for f, v in fields(self, not self.__dataclass__["hide_internals"]).items())})'
-_generated_attr = lambda self, *args: exec('raise AttributeError("Frozen class")')
+
+def __eq__(self: DataClass, other: DataClass):
+    return type(self) is type(other) and as_tuple(self) == as_tuple(other)
+
+
+def __lt__(self: DataClass, other: DataClass):
+    if isinstance(other, type(self)):
+        return as_tuple(self) < as_tuple(other)
+    return NotImplemented
+
+
+def __hash__(self):
+    return hash((type(self), *(v for v in as_tuple(self) if v is isinstance(v, Hashable))))
+
+
+def __iter__(self):
+    return iter(as_tuple(self))
+
+
+def __repr__(self):
+    show_internals = not self.__dataclass__['hide_internals']
+    field_values = ', '.join(f'{f}={v!r}' for f, v in values(self, show_internals).items())
+    return f'{type(self).__name__}({field_values})'
+
+
+def __setattr__(self, *args):
+    raise AttributeError('Frozen class')
